# sage_setup: distribution = sagemath-singular
cdef _fmat(fvars, Nk_ij, one, a, b, c, d, x, y)
cpdef _backward_subs(factory, bint flatten=*)
cpdef executor(tuple params)
cpdef _solve_for_linear_terms(factory, list eqns=*)
